#include <math.h>
#include "pedsolver.h"
#include "pglobal.h"
#include "xfile.h"
#include "mpi.h"


/**
   function solves problems of eigenvibration in parallel
   
   @param x - array for eigenmodes
   @param w - array for eigenvalues
   
   JK, 2.6.2011
*/
void parallel_solution_eigen_dynamics (double *x,double *w)
{
  long i;
  
  //  stiffness matrix
  stiffness_matrix (0);
  //  mass matrix
  mass_matrix (0);
  
  switch (Mp->eigsol.teigsol){
  case inv_iteration:{
    inverse_iteration (x,w);
    break;
  }
  case subspace_it_jacobi:{
    subspace_iter_jac (x,w);
    break;
  }
  case subspace_it_gsortho:{
    subspace_iter_ortho (x,w);
    break;
  }
  default:{
    par_print_err(Myrank+1, proc_name, "unknown eigenvalue solver is required",__FILE__, __LINE__, __func__);
  }
  }
  
  print_eigenvalues (w);
  print_eigenvectors ();
  //print_eigenvect_martin (Out);
  
  
  print_init(-1, "wt");    
  for (i=0;i<Mp->eigsol.neigv;i++){
    //  computes and prints required quantities
    compute_req_val(i);
    //compute_ipstrains (i);
    //compute_ipstresses (i);
    //compute_req_val (i);
    print_step(i, i+1, w[i], NULL);    
  }
  print_close();

}

/**
   function computes the smallest eigenvalue and appropriate eigenvector
   of the generalized problem Ax=wBx in parallel

   the eigenvector of the smallest eigenvalue is stored in Lsrs in lhs
   
   @param x - the first eigenmode
   @param w - the smallest eigenvalue
   
   JK, 2.6.2011
*/
void parallel_inverse_iteration (double *x,double *w)
{
  long i,ni,n;
  double nom,denom,zero,rho,prevrho,err,error;
  double *p,*z;
  
  //  maximum number of iterations
  ni=Mp->eigsol.nies;
  //  required error
  err=Mp->eigsol.erres;
  //  number of unknowns on appropriate subdomain
  n=Ndofm;
  //  computer zero
  zero=Mp->zero;
  

  p = new double [n];
  z = new double [n];
  
  //  initial values of the vector x
  for (i=0;i<n;i++){
    x[i]=1.0;
  }
  //  initial vector z
  Mmat->gmxv (x,z);
  
  //  main iteration loop
  for (i=0;i<ni;i++){
    //  copy of vector z which will be overwriten in back substitution
    copyv (z,p,n);
    //  back substitution
    Psol->par_linear_solver (Gtm,Smat,x,p,Out,Mespr);
    
    //  x.K.x
    nom=Psol->pss (x,z,Out);
    
    //  new local vector z
    Mmat->gmxv (x,z);

    //  x.M.x
    denom=Psol->pss(x,z,Out);
    rho=nom/denom;
    denom=1.0/sqrt(denom);
    //  normed local vector z
    cmulv (denom,z,n);

    if (Mespr==1)  printf ("\n iterace %4ld     rho %e",i,rho);
    if (i!=0){
      error=(prevrho-rho)/prevrho;
      if (error<err){
	cmulv (denom,x,n);
	break;
      }
    }
    prevrho=rho;
  }
  Mp->eigsol.anies=i;
  Mp->eigsol.aerres=error;
  w[0]=rho;
  
  delete [] z;  delete [] p;
}

/**
   function computes eigenvalues and eigenvectors
   subspace iteration with Gram-Schmidt orthonormalization is used
   the function is implemented for parallel computers
   
   @param x - array for eigenmodes
   @param w - array containing eigenvalues
   
   JK, 3.6.2011
*/
void parallel_subspace_iter_ortho (double *x,double *w)
{
  long i,j,k,l,ii,ni,n,nv,nrv;
  double err,error,maxerror,zero,nom,denom,alpha;
  double *z,*p,*wo;
  
  //  the number of unknowns on appropriate subdomain
  n=Ndofm;
  //  maximum number of iterations
  ni=Mp->eigsol.nies;
  //  required error
  err=Mp->eigsol.erres;
  //  number of required eigenvectors and eigenvalues
  nrv=Mp->eigsol.neigv;
  //  number of vectors used for iteration process
  nv=Mp->eigsol.nev;
  //  computer zero
  zero=Mp->zero;
  
  //  auxiliary local vectors
  z = new double [n];
  p = new double [n];
  wo = new double [nv];
  
  //  initial values of vectors x
  for (i=0;i<nv*n;i++){
    x[i]=1.0;
  }
  
  
  //  main iteration loop
  for (i=0;i<ni;i++){
    for (j=0;j<nv;j++){
      //  right hand side vector
      Mmat->gmxv (x+j*n,z);
      
      copyv (z,p,n);
      //  back substitution
      Psol->par_linear_solver (Gtm,Smat,x+j*n,p,Out,Mespr);
      
      //  Rayleigh' quotients
      nom=Psol->pss(x+j*n,z,Out);
      //  new local vector z
      Mmat->gmxv (x+j*n,z);

      denom=Psol->pss(x+j*n,z,Out);
      w[j]=nom/denom;
    }
    
    //  Gram-Schmidt orthonormalization
    for (j=0;j<nv;j++){
      //  new local vector z
      //  it is computed again because z contains only one vector
      //  on the other hand, x contains all approximations of eigenmodes
      Mmat->gmxv (x+j*n,z);
      
      for (k=0;k<j;k++){
	alpha=Psol->pss(x+k*n,z,Out);
	for (l=0;l<n;l++){
	  x[j*n+l]-=alpha*x[k*n+l];
	}
      }
      //  new local vector z
      //  z is computed again as was described before
      Mmat->gmxv (x+j*n,z);
      
      alpha=Psol->pss(x+j*n,z,Out);
      alpha=1.0/sqrt(alpha);
      cmulv (alpha,x+j*n,n);
    }
    
    //  convergence check
    if (i!=0){
      ii=0;  maxerror=0.0;
      for (j=0;j<nrv;j++){
	error=(wo[j]-w[j])/w[j];
	if (error>maxerror)  maxerror=error;
	if (error<err)  ii++;
      }
      
      if (Mespr==1)  printf ("\n iterace  %4ld    pocet doiterovanych tvaru %4ld  max. odchylka %e",i,ii,maxerror);
      //if (Mespr==1)  printf ("\n iterace  %4ld    w[0] %30.20e",i,w[0]);
      
      if (ii==nrv){
	//  the number of required eigenvectors are obtained
	break;
      }
    }
    
    for (j=0;j<nv;j++){
      wo[j]=w[j];
    }
  }
  
  Mp->eigsol.anies=i;
  Mp->eigsol.aerres=maxerror;
  
  delete [] z;
  delete [] wo;
  delete [] p;
}

