/*
    File:            consol_wf1.cpp
    Author:          Tomas Krejci, 15/04/2021
    Purpose:         material model for saturated-nonsaturated one-phase (water) flow in deforming medium
    sources:         Lewis and Schrefler's book pp. 86-89
    unknowns:        number of unknowns=1, pw = liqud water pressure
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "aliast.h"
#include "constrel.h"
#include "consol_wf1.h"
#include "globalt.h"
#include "globmatt.h"

con_wf1mat::con_wf1mat()
{
  compress = 0;          //compressible grains: 0=no; 1=yes
  por_type = 0;          //porosity calculation type
  krw_type = 0;          //relative permeability calculation type
  kintr_type = 0;        //intrinsic permability calculation type
  sr_type = 1;           //retention curve calculation type

  mefel_units = 1.0;     //basic units for pressures = Pa (Pascals) in MEFEL part
  vol_strain_effect = 0; //volumetric strain rate influence: 0=no; 1=yes 

  alpha = 1.0;       //Biot's constant [-] alpha = 1 - kt/ks
  ks = 2.167e9;      //bulk modulus of solid phase (grains) [Pa]
  kw = 2.0e9;        //bulk modulus of water [Pa]
  phi0 = 0.297;      //initial porosity [-]
  kintr0 = 4.5e-13;  //intrinsic permeability [m^2]
  rhow0 = 1000.0;    //initial water density [kg/m^3]
  muw0 = 1.0e-3;     //initial water viscosity [Pa.s]
  rhos0 = 2500.0;    //initial solid density

  pw_bc = 0.0;       //free boundary pressure

  sirr = 0.0;        //lower saturation degree
  ssat = 1.0;        //fully saturation degree
  lambda_krw = 1.9;  //parameter for exponential function of relative permeability

  b1 = 9.91;     //material parameter of the intrinsic permeability
  phi01 = 0.047; //reference porosity associated with kintr0
}

con_wf1mat::~con_wf1mat()
{}


/**
   function computes conductivity %matrix of the material
   in the required integration point
   
   @param d - conductivity %matrix of material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point

*/
void con_wf1mat::matcond (matrix &d,long ri,long ci,long ipp)
{
  long n;
  n = d.n;

  switch (n){
  case 1:{
    matcond1d (d,ri,ci,ipp);//1D
    break;
  }
  case 2:{
    matcond2d (d,ri,ci,ipp);//2D
    break;
  }
  case 3:{
    matcond3d (d,ri,ci,ipp);//3D
    break;
  }
  default:{
    print_err("unknown number of components of conductivity tensor is required",__FILE__,__LINE__,__func__);
  }
  }
}


/**
   function creates conductivity %matrix of the material for 1D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::matcond1d (matrix &d,long ri,long ci,long ipp)
{
  double kk;
  double pw;
  
  pw = Tm->ip[ipp].av[0];
  kk = get_kww(pw,ipp);
  
  fillm(0.0,d);

  d[0][0] = kk;
}

/**
   function creates conductivity %matrix of the material for 2D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::matcond2d (matrix &d,long ri,long ci,long ipp)
{
  double kk;
  double pw;
  
  pw = Tm->ip[ipp].av[0];
  kk = get_kww(pw,ipp);
  
  fillm(0.0,d);
  
  d[0][0] = kk;   d[0][1] = 0.0;
  d[1][0] = 0.0;  d[1][1] = kk;
}

/**
   function creates conductivity %matrix of the material for 3D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/

void con_wf1mat::matcond3d (matrix &d,long ri,long ci,long ipp)
{
  double kk;
  double pw;
  
  pw = Tm->ip[ipp].av[0];
  kk = get_kww(pw,ipp);
  
  fillm(0.0,d);
  
  d[0][0]=kk;   d[0][1]=0.0;  d[0][2]=0.0;
  d[1][0]=0.0;  d[1][1]=kk;   d[1][2]=0.0;
  d[2][0]=0.0;  d[2][1]=0.0;  d[2][2]=kk;
}


/**
   function creates capacity %matrix of the material

   @param c   - capacity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::matcap (double &cc,long ri,long ci,long ipp)
{
  double pw;
  
  pw = Tm->ip[ipp].av[0];
  cc = get_capww(pw,ipp);
}

/**
   function reads parameters
   
   @param in - input file

   15/04/2021, TKr
*/
void con_wf1mat::read(XFILE *in)
{
  
  xfscanf (in,"%k%m","waterflowtype",&waterflowtype_kwdset, &model_type); //water flow model type
  xfscanf (in,"%d", &compress);                                           //compressibility of grains
  // common material parameters
  xfscanf (in,"%le %le %le %le %le %d %d %d %d", &alpha, &ks, &rhos0, &pw_bc, &kintr0, &por_type, &kintr_type, &krw_type, &sr_type);

  switch (model_type){
  case lewis_and_schrefler:{//Lewis and Schrefler's model approach
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's model approach coupled with MEFEL
    xfscanf (in,"%le %d", &mefel_units, &vol_strain_effect);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 

  //porosity calculation:
  switch (por_type){
  case 0:{//constant
    xfscanf (in,"%le",&phi0);
    break;
  }
  case 1:{//dependent on mefel calculation
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 


  //intrinsic permebility calculation:
  switch (kintr_type){
  case 0:{//constant
    break;
  }
  case 1:{//dependent on porosity
    xfscanf (in,"%le %le", &b1, &phi01);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 

  //relative permebility calculation:
  switch (krw_type){
  case 0:{//constant
    break;
  }
  case 1:
  case 2:{//dependent on saturation degree:
    xfscanf (in,"%le %le", &sirr, &ssat);
    break;
  }
  case 3:{//exponential
    xfscanf (in,"%le", &lambda_krw);
    break;
  }
  case 4:{//liakopoulos
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 

  //retention curve type:
  switch (sr_type){
  case lewis_and_schrefler_sr:{//Lewis and Schrefler's book
    lewis_ret.read(in);
    break;
  }
  case gardner_exponential_sr:{//dependent on saturation degree:
    gardner_ret.read(in);
    break;
  }
  case potts_log_linear_sr:{//exponential
    potts_ret.read(in);
    break;
  }
  case van_genuchten_sr:{//partially saturated medium =Van Genuchten model
    van_genuchten_ret.read(in);
    break;
  }
  case van_genuchten2_sr:{//partially saturated medium =Van Genuchten model
    //van_genuchten_ret.read2(in);
    break;
  }
  case mefel_sr:{//from MEFEL
    break;
  }
  case table_sr:{//from table
    //reading of retention curve:
    data.read (in);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 

}


/**
   function prints parameters
   
   @param in - input file

   01/05/2008, TKr
*/
void con_wf1mat::print(FILE *out)
{
  /*  
      fprintf (out,"\n %d ", int(model_type));
      fprintf (out,"\n %d ", compress);
      
      switch (model_type){
      case lewis_and_schrefler:{//Lewis and Schrefler's book
      fprintf (out,"\n %le %le %le %le %le %le %le %le \n", alpha, ks, phi0, kw, rhow, muw0, kintr, rhos0);
      lewis_ret.print(out);
      break;
      }
      case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book
      fprintf (out,"\n %le %le %le %le %le %le %le %le %d %le %le %d \n", mefel_units, alpha, ks, phi0, kw, rhow, muw0, kintr, krw_type, rhos0, pw_bc, vol_strain_effect);
      break;
      }
      
      case lewis_and_schrefler_mefel_table:{//Lewis and Schrefler's book
      fprintf (out,"\n %le %le %le %le %le %le %le %le %d %le %le %d \n", mefel_units, alpha, ks, phi0, kw, rhow, muw0, kintr, krw_type, rhos0, pw_bc, vol_strain_effect);
      
      //printing of retention curve:
      data.print (out);
      fprintf (out,"\n");
      
      break;
      }
      
      case gardner_exponential:{//partially saturated medium = Exponential model, Gardner 
      gardner_ret.print(out);
      break;
      }
      case potts_log_linear:{//partially saturated medium = Log linear model, Potts
      potts_ret.print(out);
      break;
      }
      case van_genuchten:{//partially saturated medium =Van Genuchten model
      fprintf (out,"\n %le %le %le %le %le %le %le %le %le %le %d \n", mefel_units, alpha, ks, phi0, kw, rhow, muw0, kintr, rhos0, pw_bc, vol_strain_effect);
      van_genuchten_ret.print(out);
      break;
      }
      case consol_camclay:
      case consol_camclay_mech:{//fully saturated isotropic consolidation with changing porosity from cam-clay model
      fprintf (out,"\n %le %le %le %le \n", kintr, phi0, m, pw_bc);
      break;
      }
      default:{
      print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
      abort();
      }
      } 
  */
}





/**
   function computes volume part of right-hand side matrix
   in the required integration point
   
   @param d - right-hand side %matrix of material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::rhs_volume (matrix &d,long ri,long ci,long ipp)
{
  long m;
  m = d.m;
  
  switch (m){
  case 1:{
    rhs1d1 (d,ri,ci,ipp);//1D
    break;
  }
  case 2:{
    rhs2d1 (d,ri,ci,ipp);//2D
    break;
  }
  case 3:{
    rhs3d1 (d,ri,ci,ipp);//3D
    break;
  }
  default:{
    print_err ("unknown number of components of stress tensor is required", __FILE__, __LINE__, __func__);
    abort();
  }
  }
}


/**
   function computes volume part 2 of right-hand side matrix
   in the required integration point
   
   @param cc - right-hand side coefficient of material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::rhs_volume2 (double &cc,long ri,long ci,long ipp)
{
  double pw;
  
  pw = Tm->ip[ipp].av[0];
  cc = get_fwu(pw,ipp);
}



/**
   function creates volume right-hand side matrix of the material for 1D problems
   @param d - right-hand %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::rhs1d1 (matrix &d,long ri,long ci,long ipp)
{
  double f,pw;
  
  pw = Tm->ip[ipp].av[0];
  f = get_fw1(pw,ipp);
  fillm(0.0,d);
  
  d[0][0] = f*Tp->gr[0];
}

/**
   function creates volume right-hand side matrix of the material for 2D problems
   @param d - right-hand %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::rhs2d1 (matrix &d,long ri,long ci,long ipp)
{
  double f,pw;
  
  pw = Tm->ip[ipp].av[0];
  
  f = get_fw1(pw,ipp);
  fillm(0.0,d);

  d[0][0] = f*Tp->gr[0];
  d[1][0] = f*Tp->gr[1];
}

/**
   function creates volume right-hand side matrix of the material for 3D problems
   @param d - right-hand %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
*/
void con_wf1mat::rhs3d1 (matrix &d,long ri,long ci,long ipp)
{
  double f,pw;
  
  pw = Tm->ip[ipp].av[0];
  
  f = get_fw1(pw,ipp);
  fillm(0.0,d);

  d[0][0] = f*Tp->gr[0];
  d[1][0] = f*Tp->gr[1];
  d[2][0] = f*Tp->gr[2];
}




/**
   function computes degree of saturation(sorption curve)
   @param pw - water pressure

   @retval sw - degree of saturation

   15/04/2021, TKr
*/
double con_wf1mat::get_sw(double pw, long ipp)
{
  double sw;
  sw = 0.0;
  
  switch (sr_type){
  case lewis_and_schrefler_sr:{//Lewis and Schrefler's book
    sw = lewis_ret.sw(pw);
    break;
  }
  case mefel_sr:{//saturation degree ind its derivative are obtained from mefel;
    sw = Tm->givenontransq(saturation_deg, ipp); //actual saturation degree
    break;
  }
  case table_sr:{//saturation degree and its derivative are obtained from table;
    //actual saturation degree    
    if (data.tfunc == tab)
      {
	if ((pw < data.tabf->x[0]) || (pw > data.tabf->x[data.tabf->asize-1]))
	  {
	    print_err("required value %le is out of table range <%le;%le> on ip=%ld\n", 
		      __FILE__, __LINE__, __func__, pw, data.tabf->x[0], data.tabf->x[data.tabf->asize-1], ipp);
	    abort();
	  }
      }
    sw = data.getval (pw);
    
    break;
  }
    
  case gardner_exponential_sr:{//partially saturated medium = Exponential model, Gardner 
    sw = gardner_ret.sw(pw);
    break;
  }
  case potts_log_linear_sr:{//partially saturated medium = Log linear model, Potts
    sw = potts_ret.sw(pw);
    break;
  }
  case van_genuchten_sr:{//partially saturated medium = Van Genuchten model
    sw = van_genuchten_ret.sw(pw);
    break;
  }
  case van_genuchten2_sr:{//partially saturated medium = Van Genuchten model
    //sw = van_genuchten_ret.sw2(pw);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
  return(sw);
}


/**
   function computes partial derivative of degree of saturation with respect to pw, specific wter content
   @param pw - water pressure

   @retval dsw_dpw - partial derivative of degree of saturation with respect to pw

   15/04/2021, TKr
*/
double con_wf1mat::get_dsw_dpw(double pw, long ipp)
{
  double dsw_dpw;
  dsw_dpw = 0.0;

  switch (sr_type){
  case lewis_and_schrefler_sr:{//Lewis and Schrefler's book
    dsw_dpw = lewis_ret.dsw_dpw(pw);
    break;
  }
  case mefel_sr:{//saturation degree ind its derivative are obtained from mefel;
    dsw_dpw = Tm->givenontransq(der_saturation_deg, ipp); //actual derivative of saturation degree
    dsw_dpw = dsw_dpw/mefel_units; //basic units = Pa
    break;
  }

  case table_sr:{//saturation degree and its derivative are obtained from table;
    dsw_dpw = data.getderiv (pw); //actual derivative of saturation degree
    break;
  }

  case gardner_exponential_sr:{//partially saturated medium = Exponential model, Gardner 
    dsw_dpw = gardner_ret.dsw_dpw(pw);
    break;
  }
  case potts_log_linear_sr:{//partially saturated medium = Log linear model, Potts
    dsw_dpw = potts_ret.dsw_dpw(pw);
    break;
  }
  case van_genuchten_sr:{//partially saturated medium = Van Genuchten model
    dsw_dpw = van_genuchten_ret.dsw_dpw(pw);
    break;
  }
  case van_genuchten2_sr:{//partially saturated medium = Van Genuchten model
    //dsw_dpw = van_genuchten_ret.dsw_dpw2(pw);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
  return(dsw_dpw);
}


/**
   function computes water relative permeability
   @param pw - water pressure
   
   @retval krw - water relative permeability

   15/04/2021, TKr
*/
double con_wf1mat::get_krw(double pw, long ipp)
{
  double sw,krw;
  
  switch (krw_type){
  case 0:{//constant
    krw = 1.0;
    break;
  }
  case 1:{//dependent on saturation degree
    sw = get_sw(pw,ipp);
    krw = (sw-sirr)/(ssat-sirr);
    break;
  }
  case 2:{//dependent on saturation degree
    sw = get_sw(pw,ipp);
    krw = (sw-sirr)*(sw-sirr)*(sw-sirr)/((ssat-sirr)*(ssat-sirr)*(ssat-sirr));
    break;
  }
  case 3:{//exponential
    sw = get_sw(pw,ipp);
    krw = pow(sw,lambda_krw);
    break;
  }
  case 4:{//liakopoulos
    sw = get_sw(pw,ipp);
    krw = 1.0-2.207*pow((1.0-sw),1.0121);
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
  return(krw);
}


/**
   function returns porosity

   @retval phi - porosity

   15/04/2021, TKr
*/
double con_wf1mat::give_porosity(long ipp)
{
  double por;
  
  //porosity calculation:
  switch (por_type){
  case 0:{//constant
    por = phi0;
    break;
  }
  case 1:{//dependent on mefel calculation
    if(Tm->nontransq != NULL){
      por = Tm->givenontransq(porosity, ipp);// from mefel
    }
    else
      por = phi0;//for testing
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
  
  return(por);
}


/**
   function computes intrinsic permeability

   @retval kintr - intrinsic permeability

   15/04/2021, TKr
*/
double con_wf1mat::get_kintr(long ipp)
{
  double kintr,phi;

  switch (kintr_type){
  case 0:{//constant
    kintr = kintr0;
    break;
  }
  case 1:{//dependent on porosity
    phi = give_porosity(ipp);
    kintr = kintr0*exp(b1*(phi - phi01));
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 

  return(kintr);
}

/**
   function returns Biot's constant

   @retval alpha - Biot's constant
*/
double con_wf1mat::get_alpha()
{
  return(alpha);
}


/**
   function returns kw bulk modulus of water [Pa]

   @retval kw - bulk modulus of water [Pa]
*/
double con_wf1mat::get_kw()
{
  return(kw);
}


/**
   function returns muw water viscosity [Pa.s]

   @retval muw - water viscosity [Pa.s]
*/
double con_wf1mat::get_muw()
{
  return(muw0);
}


/**
   function creates conductivity coefficient of the general material
   @param pw - water pressure

   @retval kww - conductivity coefficient

   15/04/2021, TKr
*/
double con_wf1mat::get_kww(double pw, long ipp)
{
  double krw,kww,kintr;
  
  krw = get_krw(pw,ipp);
  kintr = get_kintr(ipp);
  kww = krw*kintr/muw0;//p. 89

  return(kww);
}



/**
   function creates capacity coefficient of the general material
   @param pw - water pressure

   @retval capww - capacity coefficient

   15/04/2021, TKr
*/
double con_wf1mat::get_capww(double pw, long ipp)
{
  double sw,dsw_dpw,n,capww;  
  
  sw = get_sw(pw,ipp);
  dsw_dpw = get_dsw_dpw(pw,ipp);
  n = give_porosity(ipp);
  
  if(compress == 1){
    //compressible grains:
    capww = (alpha-n)/ks*sw*(sw + dsw_dpw*pw)+n*sw/kw;
    capww = capww + n*dsw_dpw;//this equation is correct
  }
  else{
    //incompressible grains:
    capww = n*dsw_dpw;//this equation is correct
  }    
  return(capww);
}


/**
   function returns coefficient for righ-hand side of the general material
   @param pw - water pressure

   @retval fw - first part for right-hand side for continutiy equation

   15/04/2021, TKr
*/
double con_wf1mat::get_fw1(double pw, long ipp)
{
  double krw,kintr,fw1;
  
  fw1 = 0.0;
  krw = get_krw(pw,ipp);
  kintr = get_kintr(ipp);
  fw1 = krw*kintr/muw0*rhow0;//p. 89
  
  return(fw1);
}



/**
   function returns coefficient for righ-hand side of the general material
   @param pw - water pressure

   @retval fwu - volumetric strain rate effect on pore water pressure; only for partially coupled METR version

   15/04/2021, TKr
*/
double con_wf1mat::get_fwu(double pw, long ipp)
{
  double sw,fwu,depsv_r, dsr_depsv,n;
  
  fwu = 0.0;
  
  switch (model_type){
  case lewis_and_schrefler:{
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book
    if(vol_strain_effect == 1){
      n = give_porosity(ipp);
      depsv_r = Tm->givenontransq(strain_vol_rate, ipp);      //actual rate of volumetric strain from MEFEL
      Tm->ip[ipp].eqother[0] = depsv_r; //this is not necessary
      if (sr_type == mefel_sr){
	dsr_depsv = Tm->givenontransq(der_saturation_deg_depsv, ipp); //actual derivative of saturation degree with respect to volumetric strain
	Tm->ip[ipp].eqother[1] = dsr_depsv; //this is not necessary
      }
      
      sw = get_sw(pw,ipp);
      fwu = -alpha*sw*depsv_r;//volumetric strain effect
      if (sr_type == mefel_sr){
	fwu = fwu -n*dsr_depsv*depsv_r;//volumetric strain effect on retention curve
      }
    }
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  }    
  return(fwu);
}



/**
   function computes new transmission coefficient (for transmission_vector) 
   for boundary condition (third kind of boundary condition)

   @param trc     - prescribed transmission coefficient on the boundary
   @param ri      - row index
   @param ci      - column index
   @param nn      - number of node
   @param bc      - type of boundary condition
   @param ipp     - number of first integration point on element
*/
double con_wf1mat::transmission_transcoeff(double trc,long ri,long ci,long nn,long bc,long ipp)
{
  double c,pw;
  c = 0.0;
  
  pw = nodalval (nn,0);
  
  c = get_transmission_transcoeff_ww(pw,bc,ipp);

  c = c*trc;

  return (c);
}


/**
   function computes new transmission coefficient (for transmission_vector) 
   for boundary condition (third kind of boundary condition)

   @param trc     - prescribed transmission coefficient on the boundary
   @param ri      - row index
   @param ci      - column index
   @param nn      - number of node
   @param bc      - type of boundary condition
   @param ipp     - number of first integration point on element
*/
double con_wf1mat::transmission_transcoeff(double trc,long ri,long ci,long nn,long bc,long ipp, int flag)
{
  double c,pw;
  c = 0.0;
  
  pw = nodalval (nn,0);

  c = get_transmission_transcoeff_ww(pw,bc,ipp,flag);
  
  c = c*trc;
  
  return (c);
}

/**
   function computes new nodal value (for transmission_vector) 
   for boundary condition (third kind of boundary condition)

   @param nodval     - prescribed nodal value
   @param trc2       - second prescribed transmission coefficient on the boundary, 
                       if is needed (for example heat radiation coef.)
   @param ri         - row index
   @param ci         - column index
   @param nn         - number of node
   @param bc         - type of boundary condition
   @param ipp        - number of first integration point on element
*/
double con_wf1mat::transmission_nodval(double nodval,double trc2,long ri,long ci,long nn,long bc,long ipp)
{
  double c,pw;
  c = 0.0;
  
  pw = nodalval (nn,0);

  c = get_transmission_nodval_ww(nodval,pw,bc,ipp);

  return (c);
}


/**
   function computes flux (for transmission_vector) 
   for boundary condition (third kind of boundary condition)

   @param nodval     - prescribed nodal value
   @param trc2       - second prescribed transmission coefficient on the boundary, 
                       if is needed (for example heat radiation coef.)
   @param ri         - row index
   @param ci         - column index
   @param nn         - number of node
   @param bc         - type of boundary condition
   @param ipp        - number of first integration point on element
*/
double con_wf1mat::transmission_flux(double nodval,double trc2,long ri,long ci,long nn,long bc,long ipp)
{
  double c,pw;
  c = 0.0;
  
  pw = nodalval (nn,0);

  c = get_transmission_flux_ww(nodval,pw,bc,ipp);

  return (c);
}



/**
   function creates correct transfer coefficient on the boundary (transmission) for pw

   @param pw - pore water pressure on the boundary
   @param bc - type of boundary condition
   @param ipp - number of first integration point on element
*/
double con_wf1mat::get_transmission_transcoeff_ww(double pw,long bc,long ipp)
{
  double trc;
  
  //water pressure check
  waterpress_check(pw,ipp);

  //other conditions types will be added according to the neeed
  switch (bc){//type of prescribed variable
  case 30:{//pore water pressure

    trc = 1.0;
    
    break;
  }
  case 40:{//simulation of free soil surface
    if(pw < pw_bc)
      trc = 0.0;
    else
      trc = 1.0e+03;
    break;
  }
  default:{
    print_err ("no real boundary condition is prescribed", __FILE__, __LINE__, __func__);
    exit(0);
  }
  }
  return(trc);
}



/**
   function creates correct transfer coefficient on the boundary (transmission) for pw

   @param pw - pore water pressure on the boundary
   @param bc - type of boundary condition
   @param ipp - number of first integration point on element
*/
double con_wf1mat::get_transmission_transcoeff_ww(double pw,long bc,long ipp,int flag)
{
  double trc;
  
  //water pressure check
  waterpress_check(pw,ipp);

  //other conditions types will be added according to the neeed
  switch (bc){//type of prescribed variable
  case 30:{//pore water pressure
    
    trc = 1.0;
    
    break;
  }
  case 33:{//pore water pressure - flux
    if(flag == 1)
      trc=0.0;//into right hand side (matrix)
    else{
      trc=1.0;//into left hand side (flux)
    }
    break;
  }
  case 40:{//simulation of free soil surface
    if(pw < pw_bc)
      trc = 0.0;
    else
      trc = 1.0e+03;
    break;
  }
  default:{
    print_err ("no real boundary condition is prescribed", __FILE__, __LINE__, __func__);
    exit(0);
  }
  }
  return(trc);
}

/**
   function creates correct new nodal value on the boundary (transmission) for pw

   @param bv - prescribed value near the boundary
   @param pw - actual pore water pressure on the boundary
   @param bc - type of boundary condition
   @param ipp - number of first integration point on element
*/
double con_wf1mat::get_transmission_nodval_ww(double bv,double pw,long bc,long ipp)
{
  double new_nodval;
  
  //water pressure check
  waterpress_check(pw,ipp);

  switch (bc){//type of prescribed variable
  case 30:{//pore water pressure
    new_nodval = bv;
    break;
  }
  case 33:{//pore water pressure - flux
    new_nodval = bv - pw;//minus sign
    break;
  }
  case 40:{//simulation of free soil surface
    new_nodval = bv;
    break;
  }
  default:{
    print_err ("no real boundary condition is prescribed", __FILE__, __LINE__, __func__);
    exit(0);
  }
  }
  return(new_nodval);
}



/**
   function creates flux on the boundary (transmission - convective mass transfer) for pw

   @param bv - prescribed value near the boundary
   @param pw - actual pore water pressure on the boundary
   @param bc - type of boundary condition
   @param ipp - number of first integration point on element
*/
double con_wf1mat::get_transmission_flux_ww(double bv,double pw,long bc,long ipp)
{
  double flux,trc;
  
  //water pressure check
  waterpress_check(pw,ipp);

  switch (bc){//type of prescribed variable
  case 30:{//pore water pressure
    flux = bv - pw;//minus sign
    break;
  }
  case 33:{//pore water pressure - flux
    flux = bv - pw;//minus sign
    break;
  }
  case 40:{//simulation of free soil surface
    if(pw < pw_bc)
      trc = 0.0;
    else
      trc = 1.0e+03;
    flux = trc*(bv - pw);//minus sign
    break;
  }
  default:{
    print_err ("no real boundary condition is prescribed", __FILE__, __LINE__, __func__);
    exit(0);
  }
  }

  return(flux);
}




/**
   function computes all other variables at nodes
   @param compother - number of other components
   @param pw - water capillary pressure on actual node
   @param ipp - first integration point on element

   @retval other - other variable

   15/04/2021, TKr
*/

double con_wf1mat::get_othervalue(long compother,double pw, long ipp)
{
  double other;
  state_eq tt;

  switch (compother){
  case 0:{//capillary pressure
    other = -pw;
    break;
  }
  case 1:{//saturation
    other = get_sw(pw,ipp);
    break;
  }
  case 2:{//derivative of retention curve
    other = get_dsw_dpw(pw,ipp);
    break;
  }    
  case 3:{//specific moisture content
    other = get_w(pw,ipp);
    break;
  }    
  default:{
    print_err("unknown type of component is required in function", __FILE__, __LINE__, __func__);
    abort();
  }
  }
  return (other);

}



/**
     function prints names of all other variables at nodes
     @param out - output rhle
     @param compother - number of other components
*/
void con_wf1mat::print_othervalue_name(FILE *out,long compother)
{
  switch (compother){
  case 0:{//capillary pressure
    fprintf (out,"Capillary pressure (Pa)");
    break;
  }
  case 1:{//saturation
    fprintf (out,"Degree of saturation ()       ");
    break;
  }
  case 2:{//derivative of retention curve
    fprintf (out,"Derivative of retention curve (Pa^-1)      ");
    break;
  }    
  case 3:{//moisture content
    fprintf (out,"Moisture (water) content (kg/kg)      ");
    //fprintf (out,"Moisture content (kg/kg)      ");
    break;
  }    
  default:{
    print_err ("unknown type of component is required in function", __FILE__, __LINE__, __func__);
    abort();
  }
  }
}



/**
   function checks if computed unknowns are physically reasonable
   @param nv  - vector of unknowns
   @param ipp - number of integration point
*/
void con_wf1mat::values_correction (vector &nv, long ipp)
{
  //  pore water pressure control
  waterpress_check(nv[0],ipp);
}


/**
   function checks if water pressure is non-positive
   @param pw - pore water pressure
   @param ipp - number of integration point
*/
void con_wf1mat::waterpress_check(double &pw,long ipp)
{

}

/**
  This function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
*/
void con_wf1mat::updateval (long ipp)
{
}


/**
   This function initializes material model data
   with respect of consistency parameter gamma.

   @param ipp - integration point number
*/
void con_wf1mat::initval(long ipp)
{
}



/**
  The function returns ordered dof names of primary unknowns 
  required by the model.

  @param dofname   - array of uknown name for particular nodal dofs (output)
                     dofname[i] = name of i-th nodal unknown (for names see aliast.h - enum namevart)
  @param ntm       - number of transported media = number of nodal dof = length of array dofname

  Created by Tomas Krejci according to Tomas Koudelka, 28/04/2014
*/
void con_wf1mat::give_dof_names(namevart *dofname, long ntm)
{
  if (ntm < 1)
  {
    print_err("the model defines %ld unknowns but number of transported media is %ld", 
              __FILE__, __LINE__, __func__, 1, ntm);
    abort();
  }
  dofname[0] = trf_water_press;
}




/**
   function returns capillary pressure

   @param ipp - integration point number

   @retval pc - capillary water pressure
*/
double con_wf1mat::give_capillary_pressure(long ipp)
{
  double pc;

  pc = -Tm->ip[ipp].av[0];

  return(pc);
}



/**
   function returns water pressure

   @param ipp - integration point number

   @retval pw - pore water pressure
*/
double con_wf1mat::give_water_pressure(long ipp)
{
  double pw;

  pw = Tm->ip[ipp].av[0];

  return(pw);
}


/**
   function returns effective pore pressure (pressure has negative sign - mechanical convention)

   @param ipp - integration point number

   @retval pw - pore water pressure
*/
double con_wf1mat::give_effective_pore_pressure(long ipp)
{
  double pw,sw=1.0;
  
  switch (model_type){
  case lewis_and_schrefler:{//Lewis and Schrefler's book
    pw = Tm->ip[ipp].av[0];
    sw =  get_sw(pw,ipp);
    pw = -sw*pw;//returns effective pore pressure
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book, saturation degree ind its derivative are obtained from mefel;
    pw = Tm->ip[ipp].av[0];
    sw =  get_sw(pw,ipp);
    pw = -sw*pw/mefel_units;//returns effective pore pressure //corrected units for mefel //basic units = Pa  
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  }   
    return(pw);
}


/**
   function returns pore pressure (pressure has negative sign - mechanical convention)

   @param ipp - integration point number

   @retval pw - pore water pressure
*/
double con_wf1mat::give_pore_pressure(long ipp)
{
  double pw;
  
  switch (model_type){
  case lewis_and_schrefler:{//Lewis and Schrefler's book
    pw = Tm->ip[ipp].av[0];
    pw = -pw;//returns pore pressure
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book, saturation degree ind its derivative are obtained from mefel;
    pw = Tm->ip[ipp].av[0];
    pw = -pw/mefel_units;//returns pore pressure //corrected units for mefel //basic units = Pa  
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  }   
    return(pw);
}

/**
   function computes suction stress for MEFEL stress notation s = pw = -pc

   @param ipp - integration point number

   @retval suction - suction stress [Pa]

   15/04/2021, TKr
*/
double con_wf1mat::give_suction(long ipp)
{
  double pw,suction;
  
  pw = Tm->ip[ipp].av[0];

  switch (model_type){
  case lewis_and_schrefler:{//Lewis and Schrefler's book
    suction = pw;
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book, saturation degree ind its derivative are obtained from mefel;
    suction = pw;//this is correct, because capillary water pressure is negative
    
    suction = suction/mefel_units;//corrected units for mefel //basic units = Pa    
    break;
  }
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
  return(suction);
}


/**
   function returns the degree of saturation
   @param ipp - integration point number

   @retval saturation degree [-]
*/
double con_wf1mat::give_saturation_degree(long ipp)
{
  double pw,s;
  
  pw = Tm->ip[ipp].av[0];
  s = get_sw(pw,ipp);
  
  return(s);
}

/**
   function computes water content w [kg/kg]
   @param pw - water capillary pressure

   @retval w - computes water content w [kg/kg]
*/
double con_wf1mat::get_w(double pw,long ipp)
{
  double w,n,s;

  n = give_porosity(ipp);
  s = get_sw(pw,ipp);

  w = (n*s*rhow0)/(1.0 - n)/rhos0;

  return(w);
}


/**
  The funtion marks required non-transport quantities in the array antq.

  @param antq - array with flags for used material types
                antq[i] = 1 => quantity type nontransquant(i+1) is required
                antq[i] = 0 => quantity type nontransquant(i+1) is not required

  @return The function does not return anything, but it may change content of antq array.
*/
void con_wf1mat::give_reqntq(long *antq)
{
  switch (model_type){
  case lewis_and_schrefler:{//Lewis and Schrefler's book
    break;
  }
  case lewis_and_schrefler_mefel:{//Lewis and Schrefler's book, saturation degree ind its derivative are obtained from mefel;
    antq[porosity-1] = 1;
    if (sr_type == mefel_sr){
      antq[saturation_deg-1] = 1;
      antq[der_saturation_deg-1] = 1;
      antq[der_saturation_deg_depsv-1] = 1;
    }
    if(vol_strain_effect == 1)
      antq[strain_vol_rate-1] = 1;
    break;
  }  
  default:{
    print_err ("unknown model type is required", __FILE__, __LINE__, __func__);
    abort();
  }
  } 
}
